#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#ifndef SIGSTKSZ
# define SIGSTKSZ 16384
#endif
#ifndef STDERR_FILENO
# define STDERR_FILENO 2
#endif

static void die (int) __attribute__ ((noreturn));
static void
die (int signo)
{
  char const *message;
  message = signo ? "program error" : "stack overflow";
  write (STDERR_FILENO, message, strlen (message));
  write (STDERR_FILENO, "\n", 1);
  if (! signo)
    _exit (1);
  raise (signo);
  abort ();
}

/* Storage for the alternate signal stack.  */
static union
{
  char buffer[SIGSTKSZ];

  /* These other members are for proper alignment.  There's no
     standard way to guarantee stack alignment, but this seems enough
     in practice.  */
  long double ld;
  long l;
  void *p;
} alternate_signal_stack;

int
c_stack_action ()
{
  int r;
  stack_t st;
  struct sigaction act;

  st.ss_flags = 0;
  st.ss_sp = alternate_signal_stack.buffer;
  st.ss_size = sizeof alternate_signal_stack.buffer;
  r = sigaltstack (&st, NULL);
  if (r != 0)
    return r;

  sigemptyset (&act.sa_mask);
  act.sa_flags = SA_NODEFER | SA_RESETHAND;
  act.sa_handler = die;

  return sigaction (SIGSEGV, &act, NULL);
}

static volatile int *
recurse_1 (volatile int n, volatile int *p)
{
  if (n >= 0)
    *recurse_1 (n + 1, p) += n;
  return p;
}

static int
recurse (volatile int n)
{
  int sum = 0;
  return *recurse_1 (n, &sum);
}

int
main (int argc, char **argv)
{
  struct rlimit rl;
  rl.rlim_cur = rl.rlim_max = 0x100000; /* 1 MB */
  setrlimit (RLIMIT_STACK, &rl);

  if (c_stack_action () == 0)
    {
      return recurse (0);
    }
  fputs ("skipping test: ", stderr);
  perror ("c_stack_action");
  return 77;
}
